import { TextService, SHAPE, Camera, ContextService, SceneGraphService, Renderable, getEuler, fromRotationTranslationScale, world, RenderingPluginContribution } from '@antv/g';
import { Component } from '@antv/g-ecs';
import { detect } from 'detect-browser';

var NAMED_TAG = "named";
var INJECT_TAG = "inject";
var TAGGED = "inversify:tagged";
var TAGGED_PROP = "inversify:tagged_props";
var PARAM_TYPES = "inversify:paramtypes";
var DESIGN_PARAM_TYPES = "design:paramtypes";

var idCounter = 0;
function id() {
    return idCounter++;
}

var DUPLICATED_INJECTABLE_DECORATOR = "Cannot apply @injectable decorator multiple times.";
var DUPLICATED_METADATA = "Metadata key was used more than once in a parameter:";
var UNDEFINED_INJECT_ANNOTATION = function (name) {
    return "@inject called with undefined this could mean that the class " + name + " has " +
        "a circular dependency problem. You can use a LazyServiceIdentifer to  " +
        "overcome this limitation.";
};
var INVALID_DECORATOR_OPERATION = "The @inject @multiInject @tagged and @named decorators " +
    "must be applied to the parameters of a class constructor or a class property.";

var Metadata = (function () {
    function Metadata(key, value) {
        this.key = key;
        this.value = value;
    }
    Metadata.prototype.toString = function () {
        if (this.key === NAMED_TAG) {
            return "named: " + this.value.toString() + " ";
        }
        else {
            return "tagged: { key:" + this.key.toString() + ", value: " + this.value + " }";
        }
    };
    return Metadata;
}());

function tagParameter(annotationTarget, propertyName, parameterIndex, metadata) {
    var metadataKey = TAGGED;
    _tagParameterOrProperty(metadataKey, annotationTarget, propertyName, metadata, parameterIndex);
}
function tagProperty(annotationTarget, propertyName, metadata) {
    var metadataKey = TAGGED_PROP;
    _tagParameterOrProperty(metadataKey, annotationTarget.constructor, propertyName, metadata);
}
function _tagParameterOrProperty(metadataKey, annotationTarget, propertyName, metadata, parameterIndex) {
    var paramsOrPropertiesMetadata = {};
    var isParameterDecorator = (typeof parameterIndex === "number");
    var key = (parameterIndex !== undefined && isParameterDecorator) ? parameterIndex.toString() : propertyName;
    if (isParameterDecorator && propertyName !== undefined) {
        throw new Error(INVALID_DECORATOR_OPERATION);
    }
    if (Reflect.hasOwnMetadata(metadataKey, annotationTarget)) {
        paramsOrPropertiesMetadata = Reflect.getMetadata(metadataKey, annotationTarget);
    }
    var paramOrPropertyMetadata = paramsOrPropertiesMetadata[key];
    if (!Array.isArray(paramOrPropertyMetadata)) {
        paramOrPropertyMetadata = [];
    }
    else {
        for (var _i = 0, paramOrPropertyMetadata_1 = paramOrPropertyMetadata; _i < paramOrPropertyMetadata_1.length; _i++) {
            var m = paramOrPropertyMetadata_1[_i];
            if (m.key === metadata.key) {
                throw new Error(DUPLICATED_METADATA + " " + m.key.toString());
            }
        }
    }
    paramOrPropertyMetadata.push(metadata);
    paramsOrPropertiesMetadata[key] = paramOrPropertyMetadata;
    Reflect.defineMetadata(metadataKey, paramsOrPropertiesMetadata, annotationTarget);
}

function inject(serviceIdentifier) {
    return function (target, targetKey, index) {
        if (serviceIdentifier === undefined) {
            throw new Error(UNDEFINED_INJECT_ANNOTATION(target.name));
        }
        var metadata = new Metadata(INJECT_TAG, serviceIdentifier);
        if (typeof index === "number") {
            tagParameter(target, targetKey, index, metadata);
        }
        else {
            tagProperty(target, targetKey, metadata);
        }
    };
}

var __spreadArray = (undefined && undefined.__spreadArray) || function (to, from) {
    for (var i = 0, il = from.length, j = to.length; i < il; i++, j++)
        to[j] = from[i];
    return to;
};

var __spreadArray$1 = (undefined && undefined.__spreadArray) || function (to, from) {
    for (var i = 0, il = from.length, j = to.length; i < il; i++, j++)
        to[j] = from[i];
    return to;
};

var __awaiter = (undefined && undefined.__awaiter) || function (thisArg, _arguments, P, generator) {
    function adopt(value) { return value instanceof P ? value : new P(function (resolve) { resolve(value); }); }
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : adopt(result.value).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
var __generator = (undefined && undefined.__generator) || function (thisArg, body) {
    var _ = { label: 0, sent: function() { if (t[0] & 1) throw t[1]; return t[1]; }, trys: [], ops: [] }, f, y, t, g;
    return g = { next: verb(0), "throw": verb(1), "return": verb(2) }, typeof Symbol === "function" && (g[Symbol.iterator] = function() { return this; }), g;
    function verb(n) { return function (v) { return step([n, v]); }; }
    function step(op) {
        if (f) throw new TypeError("Generator is already executing.");
        while (_) try {
            if (f = 1, y && (t = op[0] & 2 ? y["return"] : op[0] ? y["throw"] || ((t = y["return"]) && t.call(y), 0) : y.next) && !(t = t.call(y, op[1])).done) return t;
            if (y = 0, t) op = [op[0] & 2, t.value];
            switch (op[0]) {
                case 0: case 1: t = op; break;
                case 4: _.label++; return { value: op[1], done: false };
                case 5: _.label++; y = op[1]; op = [0]; continue;
                case 7: op = _.ops.pop(); _.trys.pop(); continue;
                default:
                    if (!(t = _.trys, t = t.length > 0 && t[t.length - 1]) && (op[0] === 6 || op[0] === 2)) { _ = 0; continue; }
                    if (op[0] === 3 && (!t || (op[1] > t[0] && op[1] < t[3]))) { _.label = op[1]; break; }
                    if (op[0] === 6 && _.label < t[1]) { _.label = t[1]; t = op; break; }
                    if (t && _.label < t[2]) { _.label = t[2]; _.ops.push(op); break; }
                    if (t[2]) _.ops.pop();
                    _.trys.pop(); continue;
            }
            op = body.call(thisArg, _);
        } catch (e) { op = [6, e]; y = 0; } finally { f = t = 0; }
        if (op[0] & 5) throw op[1]; return { value: op[0] ? op[1] : void 0, done: true };
    }
};
var __spreadArray$2 = (undefined && undefined.__spreadArray) || function (to, from) {
    for (var i = 0, il = from.length, j = to.length; i < il; i++, j++)
        to[j] = from[i];
    return to;
};

var ContainerModule = (function () {
    function ContainerModule(registry) {
        this.id = id();
        this.registry = registry;
    }
    return ContainerModule;
}());

function injectable() {
    return function (target) {
        if (Reflect.hasOwnMetadata(PARAM_TYPES, target)) {
            throw new Error(DUPLICATED_INJECTABLE_DECORATOR);
        }
        var types = Reflect.getMetadata(DESIGN_PARAM_TYPES, target) || [];
        Reflect.defineMetadata(PARAM_TYPES, types, target);
        return target;
    };
}

function _classCallCheck(instance, Constructor) {
  if (!(instance instanceof Constructor)) {
    throw new TypeError("Cannot call a class as a function");
  }
}

function _defineProperties(target, props) {
  for (var i = 0; i < props.length; i++) {
    var descriptor = props[i];
    descriptor.enumerable = descriptor.enumerable || false;
    descriptor.configurable = true;
    if ("value" in descriptor) descriptor.writable = true;
    Object.defineProperty(target, descriptor.key, descriptor);
  }
}

function _createClass(Constructor, protoProps, staticProps) {
  if (protoProps) _defineProperties(Constructor.prototype, protoProps);
  if (staticProps) _defineProperties(Constructor, staticProps);
  return Constructor;
}

function _defineProperty(obj, key, value) {
  if (key in obj) {
    Object.defineProperty(obj, key, {
      value: value,
      enumerable: true,
      configurable: true,
      writable: true
    });
  } else {
    obj[key] = value;
  }

  return obj;
}

function _inherits(subClass, superClass) {
  if (typeof superClass !== "function" && superClass !== null) {
    throw new TypeError("Super expression must either be null or a function");
  }

  subClass.prototype = Object.create(superClass && superClass.prototype, {
    constructor: {
      value: subClass,
      writable: true,
      configurable: true
    }
  });
  if (superClass) _setPrototypeOf(subClass, superClass);
}

function _getPrototypeOf(o) {
  _getPrototypeOf = Object.setPrototypeOf ? Object.getPrototypeOf : function _getPrototypeOf(o) {
    return o.__proto__ || Object.getPrototypeOf(o);
  };
  return _getPrototypeOf(o);
}

function _setPrototypeOf(o, p) {
  _setPrototypeOf = Object.setPrototypeOf || function _setPrototypeOf(o, p) {
    o.__proto__ = p;
    return o;
  };

  return _setPrototypeOf(o, p);
}

function _isNativeReflectConstruct() {
  if (typeof Reflect === "undefined" || !Reflect.construct) return false;
  if (Reflect.construct.sham) return false;
  if (typeof Proxy === "function") return true;

  try {
    Boolean.prototype.valueOf.call(Reflect.construct(Boolean, [], function () {}));
    return true;
  } catch (e) {
    return false;
  }
}

function _assertThisInitialized(self) {
  if (self === void 0) {
    throw new ReferenceError("this hasn't been initialised - super() hasn't been called");
  }

  return self;
}

function _possibleConstructorReturn(self, call) {
  if (call && (typeof call === "object" || typeof call === "function")) {
    return call;
  }

  return _assertThisInitialized(self);
}

function _createSuper(Derived) {
  var hasNativeReflectConstruct = _isNativeReflectConstruct();

  return function _createSuperInternal() {
    var Super = _getPrototypeOf(Derived),
        result;

    if (hasNativeReflectConstruct) {
      var NewTarget = _getPrototypeOf(this).constructor;

      result = Reflect.construct(Super, arguments, NewTarget);
    } else {
      result = Super.apply(this, arguments);
    }

    return _possibleConstructorReturn(this, result);
  };
}

function _slicedToArray(arr, i) {
  return _arrayWithHoles(arr) || _iterableToArrayLimit(arr, i) || _unsupportedIterableToArray(arr, i) || _nonIterableRest();
}

function _toConsumableArray(arr) {
  return _arrayWithoutHoles(arr) || _iterableToArray(arr) || _unsupportedIterableToArray(arr) || _nonIterableSpread();
}

function _arrayWithoutHoles(arr) {
  if (Array.isArray(arr)) return _arrayLikeToArray(arr);
}

function _arrayWithHoles(arr) {
  if (Array.isArray(arr)) return arr;
}

function _iterableToArray(iter) {
  if (typeof Symbol !== "undefined" && iter[Symbol.iterator] != null || iter["@@iterator"] != null) return Array.from(iter);
}

function _iterableToArrayLimit(arr, i) {
  var _i = arr == null ? null : typeof Symbol !== "undefined" && arr[Symbol.iterator] || arr["@@iterator"];

  if (_i == null) return;
  var _arr = [];
  var _n = true;
  var _d = false;

  var _s, _e;

  try {
    for (_i = _i.call(arr); !(_n = (_s = _i.next()).done); _n = true) {
      _arr.push(_s.value);

      if (i && _arr.length === i) break;
    }
  } catch (err) {
    _d = true;
    _e = err;
  } finally {
    try {
      if (!_n && _i["return"] != null) _i["return"]();
    } finally {
      if (_d) throw _e;
    }
  }

  return _arr;
}

function _unsupportedIterableToArray(o, minLen) {
  if (!o) return;
  if (typeof o === "string") return _arrayLikeToArray(o, minLen);
  var n = Object.prototype.toString.call(o).slice(8, -1);
  if (n === "Object" && o.constructor) n = o.constructor.name;
  if (n === "Map" || n === "Set") return Array.from(o);
  if (n === "Arguments" || /^(?:Ui|I)nt(?:8|16|32)(?:Clamped)?Array$/.test(n)) return _arrayLikeToArray(o, minLen);
}

function _arrayLikeToArray(arr, len) {
  if (len == null || len > arr.length) len = arr.length;

  for (var i = 0, arr2 = new Array(len); i < len; i++) arr2[i] = arr[i];

  return arr2;
}

function _nonIterableSpread() {
  throw new TypeError("Invalid attempt to spread non-iterable instance.\nIn order to be iterable, non-array objects must have a [Symbol.iterator]() method.");
}

function _nonIterableRest() {
  throw new TypeError("Invalid attempt to destructure non-iterable instance.\nIn order to be iterable, non-array objects must have a [Symbol.iterator]() method.");
}

var ElementSVG = /*#__PURE__*/function (_Component) {
  _inherits(ElementSVG, _Component);

  var _super = _createSuper(ElementSVG);

  function ElementSVG() {
    var _this;

    _classCallCheck(this, ElementSVG);

    for (var _len = arguments.length, args = new Array(_len), _key = 0; _key < _len; _key++) {
      args[_key] = arguments[_key];
    }

    _this = _super.call.apply(_super, [this].concat(args));
    _this.$el = void 0;
    _this.$groupEl = void 0;
    return _this;
  }

  return ElementSVG;
}(Component);
ElementSVG.tag = 'c-svg-element';

/*! *****************************************************************************
Copyright (c) Microsoft Corporation.

Permission to use, copy, modify, and/or distribute this software for any
purpose with or without fee is hereby granted.

THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES WITH
REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY
AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY SPECIAL, DIRECT,
INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM
LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR
OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR
PERFORMANCE OF THIS SOFTWARE.
***************************************************************************** */

function __decorate(decorators, target, key, desc) {
    var c = arguments.length, r = c < 3 ? target : desc === null ? desc = Object.getOwnPropertyDescriptor(target, key) : desc, d;
    if (typeof Reflect === "object" && typeof Reflect.decorate === "function") r = Reflect.decorate(decorators, target, key, desc);
    else for (var i = decorators.length - 1; i >= 0; i--) if (d = decorators[i]) r = (c < 3 ? d(r) : c > 3 ? d(target, key, r) : d(target, key)) || r;
    return c > 3 && r && Object.defineProperty(target, key, r), r;
}

function __metadata(metadataKey, metadataValue) {
    if (typeof Reflect === "object" && typeof Reflect.metadata === "function") return Reflect.metadata(metadataKey, metadataValue);
}

var toString = {}.toString;
var isType = function (value, type) { return toString.call(value) === '[object ' + type + ']'; };

var isArray = (function (value) {
    return Array.isArray ?
        Array.isArray(value) :
        isType(value, 'Array');
});

var isString = (function (str) {
    return isType(str, 'String');
});

function parseRadius(radius) {
  var r1 = 0;
  var r2 = 0;
  var r3 = 0;
  var r4 = 0;

  if (isArray(radius)) {
    if (radius.length === 1) {
      r1 = r2 = r3 = r4 = radius[0];
    } else if (radius.length === 2) {
      r1 = r3 = radius[0];
      r2 = r4 = radius[1];
    } else if (radius.length === 3) {
      r1 = radius[0];
      r2 = r4 = radius[1];
      r3 = radius[2];
    } else {
      r1 = radius[0];
      r2 = radius[1];
      r3 = radius[2];
      r4 = radius[3];
    }
  } else {
    r1 = r2 = r3 = r4 = radius;
  }

  return {
    r1: r1,
    r2: r2,
    r3: r3,
    r4: r4
  };
}

var RectRenderer = /*#__PURE__*/function () {
  function RectRenderer() {
    _classCallCheck(this, RectRenderer);
  }

  _createClass(RectRenderer, [{
    key: "apply",
    value: function apply($el, attributes) {
      var _attributes$radius = attributes.radius,
          radius = _attributes$radius === void 0 ? 0 : _attributes$radius,
          _attributes$width = attributes.width,
          width = _attributes$width === void 0 ? 0 : _attributes$width,
          _attributes$height = attributes.height,
          height = _attributes$height === void 0 ? 0 : _attributes$height;
      var d = '';

      if (!radius) {
        d = "M 0, 0 l ".concat(width, ",0 l 0,").concat(height, " l").concat(-width, " 0 z");
      } else {
        var r = parseRadius(radius);

        if (isArray(radius)) {
          if (radius.length === 1) {
            r.r1 = r.r2 = r.r3 = r.r4 = radius[0];
          } else if (radius.length === 2) {
            r.r1 = r.r3 = radius[0];
            r.r2 = r.r4 = radius[1];
          } else if (radius.length === 3) {
            r.r1 = radius[0];
            r.r2 = r.r4 = radius[1];
            r.r3 = radius[2];
          } else {
            r.r1 = radius[0];
            r.r2 = radius[1];
            r.r3 = radius[2];
            r.r4 = radius[3];
          }
        } else {
          r.r1 = r.r2 = r.r3 = r.r4 = radius;
        }

        d = [["M ".concat(r.r1, ",0")], ["l ".concat(width - r.r1 - r.r2, ",0")], ["a ".concat(r.r2, ",").concat(r.r2, ",0,0,1,").concat(r.r2, ",").concat(r.r2)], ["l 0,".concat(height - r.r2 - r.r3)], ["a ".concat(r.r3, ",").concat(r.r3, ",0,0,1,").concat(-r.r3, ",").concat(r.r3)], ["l ".concat(r.r3 + r.r4 - width, ",0")], ["a ".concat(r.r4, ",").concat(r.r4, ",0,0,1,").concat(-r.r4, ",").concat(-r.r4)], ["l 0,".concat(r.r4 + r.r1 - height)], ["a ".concat(r.r1, ",").concat(r.r1, ",0,0,1,").concat(r.r1, ",").concat(-r.r1)], ['z']].join(' ');
      }

      $el.setAttribute('d', d);
    }
  }]);

  return RectRenderer;
}();

RectRenderer = __decorate([injectable()], RectRenderer);

var ImageRenderer = /*#__PURE__*/function () {
  function ImageRenderer() {
    _classCallCheck(this, ImageRenderer);
  }

  _createClass(ImageRenderer, [{
    key: "apply",
    value: function apply($el, attributes) {
      var img = attributes.img,
          _attributes$width = attributes.width,
          width = _attributes$width === void 0 ? 0 : _attributes$width,
          _attributes$height = attributes.height,
          height = _attributes$height === void 0 ? 0 : _attributes$height,
          _attributes$anchor = attributes.anchor,
          anchor = _attributes$anchor === void 0 ? [0, 0] : _attributes$anchor; // set anchor

      $el.setAttribute('x', "".concat(-anchor[0] * width));
      $el.setAttribute('y', "".concat(-anchor[1] * height));

      if (isString(img)) {
        $el.setAttribute('href', img);
      } else if (img instanceof Image) {
        if (!width) {
          $el.setAttribute('width', "".concat(img.width)); // TODO: set renderable.aabbDirty
          // this.attr('width', img.width);
        }

        if (!height) {
          $el.setAttribute('height', "".concat(img.height)); // this.attr('height', img.height);
        }

        $el.setAttribute('href', img.src); // @ts-ignore
      } else if (img instanceof HTMLElement && isString(img.nodeName) && img.nodeName.toUpperCase() === 'CANVAS') {
        $el.setAttribute('href', img.toDataURL()); // @ts-ignore
      } else if (img instanceof ImageData) {
        var canvas = document.createElement('canvas'); // @ts-ignore

        canvas.setAttribute('width', "".concat(img.width)); // @ts-ignore

        canvas.setAttribute('height', "".concat(img.height));
        var context = canvas.getContext('2d');

        if (context) {
          context.putImageData(img, 0, 0);

          if (!width) {
            // @ts-ignore
            $el.setAttribute('width', "".concat(img.width)); // this.attr('width', img.width);
          }

          if (!height) {
            // @ts-ignore
            $el.setAttribute('height', "".concat(img.height)); // this.attr('height', img.height);
          }

          $el.setAttribute('href', canvas.toDataURL());
        }
      }
    }
  }]);

  return ImageRenderer;
}();

ImageRenderer = __decorate([injectable()], ImageRenderer);

var LineRenderer = /*#__PURE__*/function () {
  function LineRenderer() {
    _classCallCheck(this, LineRenderer);
  }

  _createClass(LineRenderer, [{
    key: "apply",
    value: function apply($el, attributes) {
      var x1 = attributes.x1,
          y1 = attributes.y1,
          x2 = attributes.x2,
          y2 = attributes.y2,
          _attributes$x = attributes.x,
          x = _attributes$x === void 0 ? 0 : _attributes$x,
          _attributes$y = attributes.y,
          y = _attributes$y === void 0 ? 0 : _attributes$y;
      $el.setAttribute('x1', "".concat(x1 - x));
      $el.setAttribute('y1', "".concat(y1 - y));
      $el.setAttribute('x2', "".concat(x2 - x));
      $el.setAttribute('y2', "".concat(y2 - y));
    }
  }]);

  return LineRenderer;
}();

LineRenderer = __decorate([injectable()], LineRenderer);

var PolylineRenderer = /*#__PURE__*/function () {
  function PolylineRenderer() {
    _classCallCheck(this, PolylineRenderer);
  }

  _createClass(PolylineRenderer, [{
    key: "apply",
    value: function apply($el, attributes) {
      // @ts-ignore
      var points = attributes.points,
          _attributes$x = attributes.x,
          x = _attributes$x === void 0 ? 0 : _attributes$x,
          _attributes$y = attributes.y,
          y = _attributes$y === void 0 ? 0 : _attributes$y;

      if (points && points.length >= 2) {
        $el.setAttribute('points', points.map(function (point) {
          return "".concat(point[0] - x, ",").concat(point[1] - y);
        }).join(' '));
      }
    }
  }]);

  return PolylineRenderer;
}();

PolylineRenderer = __decorate([injectable()], PolylineRenderer);

var BASELINE_MAP = {
  top: 'before-edge',
  middle: 'central',
  bottom: 'after-edge',
  alphabetic: 'alphabetic',
  ideographic: 'ideographic',
  hanging: 'hanging'
}; // for FireFox
// @see https://github.com/plouc/nivo/issues/164

var BASELINE_MAP_FOR_FIREFOX = {
  top: 'text-before-edge',
  middle: 'central',
  bottom: 'text-after-edge',
  alphabetic: 'alphabetic',
  ideographic: 'ideographic',
  hanging: 'hanging'
};
var ANCHOR_MAP = {
  left: 'left',
  start: 'left',
  center: 'middle',
  right: 'end',
  end: 'end'
};

var TextRenderer = /*#__PURE__*/function () {
  function TextRenderer() {
    _classCallCheck(this, TextRenderer);

    this.textService = void 0;
  }

  _createClass(TextRenderer, [{
    key: "apply",
    value: function apply($el, attributes) {
      var textAlign = attributes.textAlign,
          text = attributes.text,
          textBaseline = attributes.textBaseline,
          _attributes$fontSize = attributes.fontSize,
          lineCap = attributes.lineCap,
          lineJoin = attributes.lineJoin,
          _attributes$lineWidth = attributes.lineWidth,
          lineWidth = _attributes$lineWidth === void 0 ? 0 : _attributes$lineWidth;
      var browser = detect();

      if (browser && browser.name === 'firefox') {
        // compatible with FireFox browser, ref: https://github.com/antvis/g/issues/119
        $el.setAttribute('dominant-baseline', BASELINE_MAP_FOR_FIREFOX[textBaseline] || 'alphabetic');
      } else {
        $el.setAttribute('dominant-baseline', BASELINE_MAP_FOR_FIREFOX[textBaseline]);
        $el.setAttribute('alignment-baseline', BASELINE_MAP[textBaseline]);
      }

      $el.setAttribute('text-anchor', ANCHOR_MAP[textAlign]);
      $el.setAttribute('paint-order', 'stroke');
      $el.setAttribute('style', "stroke-linecap:".concat(lineCap, "; stroke-linejoin:").concat(lineJoin, ";"));

      var _this$textService$mea = this.textService.measureText(text, attributes),
          lines = _this$textService$mea.lines,
          lineHeight = _this$textService$mea.lineHeight,
          height = _this$textService$mea.height;

      var lineNum = lines.length;

      if (lineNum === 1) {
        $el.innerHTML = lines[0];
        $el.setAttribute('dx', "".concat(lineWidth / 2));
      } else {
        $el.innerHTML = lines.map(function (line, i) {
          var dx = lineWidth / 2;
          var dy = 0;

          if (i === 0) {
            // TODO: handle other textBaseline values
            if (textBaseline === 'middle') {
              dy = lineHeight / 2 - height / 2;
            } else if (textBaseline === 'top' || textBaseline === 'hanging') {
              dy = 0;
            } else if (textBaseline === 'bottom' || textBaseline === 'alphabetic' || textBaseline === 'ideographic') {
              dy = -lineHeight * (lineNum - 1);
            }
          } else {
            dy = lineHeight;
          }

          return "<tspan x=\"0\" dx=\"".concat(dx, "\" dy=\"").concat(dy, "\">").concat(line, "</tspan>");
        }).join('');
      }
    }
  }]);

  return TextRenderer;
}();

__decorate([inject(TextService), __metadata("design:type", TextService)], TextRenderer.prototype, "textService", void 0);

TextRenderer = __decorate([injectable()], TextRenderer);

var PathRenderer = /*#__PURE__*/function () {
  function PathRenderer() {
    _classCallCheck(this, PathRenderer);
  }

  _createClass(PathRenderer, [{
    key: "apply",
    value: function apply($el, attributes) {
      // @ts-ignore
      var path = attributes.path,
          x = attributes.x,
          y = attributes.y;
      $el.setAttribute('d', this.formatPath(path)); // FIXME offset xy
    }
  }, {
    key: "formatPath",
    value: function formatPath(value) {
      var newValue = value.map(function (path) {
        return path.join(' ');
      }).join('');

      if (~newValue.indexOf('NaN')) {
        return '';
      }

      return newValue;
    }
  }]);

  return PathRenderer;
}();

PathRenderer = __decorate([injectable()], PathRenderer);

var ElementRendererFactory = Symbol('ElementRendererFactory');

/**
 * Common utilities
 * @module glMatrix
 */
// Configuration Constants
var EPSILON = 0.000001;
var ARRAY_TYPE = typeof Float32Array !== 'undefined' ? Float32Array : Array;
if (!Math.hypot) Math.hypot = function () {
  var y = 0,
      i = arguments.length;

  while (i--) {
    y += arguments[i] * arguments[i];
  }

  return Math.sqrt(y);
};

/**
 * 3x3 Matrix
 * @module mat3
 */

/**
 * Creates a new identity mat3
 *
 * @returns {mat3} a new 3x3 matrix
 */

function create() {
  var out = new ARRAY_TYPE(9);

  if (ARRAY_TYPE != Float32Array) {
    out[1] = 0;
    out[2] = 0;
    out[3] = 0;
    out[5] = 0;
    out[6] = 0;
    out[7] = 0;
  }

  out[0] = 1;
  out[4] = 1;
  out[8] = 1;
  return out;
}

/**
 * Returns the translation vector component of a transformation
 *  matrix. If a matrix is built with fromRotationTranslation,
 *  the returned vector will be the same as the translation vector
 *  originally supplied.
 * @param  {vec3} out Vector to receive translation component
 * @param  {ReadonlyMat4} mat Matrix to be decomposed (input)
 * @return {vec3} out
 */

function getTranslation(out, mat) {
  out[0] = mat[12];
  out[1] = mat[13];
  out[2] = mat[14];
  return out;
}
/**
 * Returns the scaling factor component of a transformation
 *  matrix. If a matrix is built with fromRotationTranslationScale
 *  with a normalized Quaternion paramter, the returned vector will be
 *  the same as the scaling vector
 *  originally supplied.
 * @param  {vec3} out Vector to receive scaling factor component
 * @param  {ReadonlyMat4} mat Matrix to be decomposed (input)
 * @return {vec3} out
 */

function getScaling(out, mat) {
  var m11 = mat[0];
  var m12 = mat[1];
  var m13 = mat[2];
  var m21 = mat[4];
  var m22 = mat[5];
  var m23 = mat[6];
  var m31 = mat[8];
  var m32 = mat[9];
  var m33 = mat[10];
  out[0] = Math.hypot(m11, m12, m13);
  out[1] = Math.hypot(m21, m22, m23);
  out[2] = Math.hypot(m31, m32, m33);
  return out;
}
/**
 * Returns a quaternion representing the rotational component
 *  of a transformation matrix. If a matrix is built with
 *  fromRotationTranslation, the returned quaternion will be the
 *  same as the quaternion originally supplied.
 * @param {quat} out Quaternion to receive the rotation component
 * @param {ReadonlyMat4} mat Matrix to be decomposed (input)
 * @return {quat} out
 */

function getRotation(out, mat) {
  var scaling = new ARRAY_TYPE(3);
  getScaling(scaling, mat);
  var is1 = 1 / scaling[0];
  var is2 = 1 / scaling[1];
  var is3 = 1 / scaling[2];
  var sm11 = mat[0] * is1;
  var sm12 = mat[1] * is2;
  var sm13 = mat[2] * is3;
  var sm21 = mat[4] * is1;
  var sm22 = mat[5] * is2;
  var sm23 = mat[6] * is3;
  var sm31 = mat[8] * is1;
  var sm32 = mat[9] * is2;
  var sm33 = mat[10] * is3;
  var trace = sm11 + sm22 + sm33;
  var S = 0;

  if (trace > 0) {
    S = Math.sqrt(trace + 1.0) * 2;
    out[3] = 0.25 * S;
    out[0] = (sm23 - sm32) / S;
    out[1] = (sm31 - sm13) / S;
    out[2] = (sm12 - sm21) / S;
  } else if (sm11 > sm22 && sm11 > sm33) {
    S = Math.sqrt(1.0 + sm11 - sm22 - sm33) * 2;
    out[3] = (sm23 - sm32) / S;
    out[0] = 0.25 * S;
    out[1] = (sm12 + sm21) / S;
    out[2] = (sm31 + sm13) / S;
  } else if (sm22 > sm33) {
    S = Math.sqrt(1.0 + sm22 - sm11 - sm33) * 2;
    out[3] = (sm31 - sm13) / S;
    out[0] = (sm12 + sm21) / S;
    out[1] = 0.25 * S;
    out[2] = (sm23 + sm32) / S;
  } else {
    S = Math.sqrt(1.0 + sm33 - sm11 - sm22) * 2;
    out[3] = (sm12 - sm21) / S;
    out[0] = (sm31 + sm13) / S;
    out[1] = (sm23 + sm32) / S;
    out[2] = 0.25 * S;
  }

  return out;
}

/**
 * 3 Dimensional Vector
 * @module vec3
 */

/**
 * Creates a new, empty vec3
 *
 * @returns {vec3} a new 3D vector
 */

function create$1() {
  var out = new ARRAY_TYPE(3);

  if (ARRAY_TYPE != Float32Array) {
    out[0] = 0;
    out[1] = 0;
    out[2] = 0;
  }

  return out;
}
/**
 * Calculates the length of a vec3
 *
 * @param {ReadonlyVec3} a vector to calculate length of
 * @returns {Number} length of a
 */

function length(a) {
  var x = a[0];
  var y = a[1];
  var z = a[2];
  return Math.hypot(x, y, z);
}
/**
 * Creates a new vec3 initialized with the given values
 *
 * @param {Number} x X component
 * @param {Number} y Y component
 * @param {Number} z Z component
 * @returns {vec3} a new 3D vector
 */

function fromValues(x, y, z) {
  var out = new ARRAY_TYPE(3);
  out[0] = x;
  out[1] = y;
  out[2] = z;
  return out;
}
/**
 * Normalize a vec3
 *
 * @param {vec3} out the receiving vector
 * @param {ReadonlyVec3} a vector to normalize
 * @returns {vec3} out
 */

function normalize(out, a) {
  var x = a[0];
  var y = a[1];
  var z = a[2];
  var len = x * x + y * y + z * z;

  if (len > 0) {
    //TODO: evaluate use of glm_invsqrt here?
    len = 1 / Math.sqrt(len);
  }

  out[0] = a[0] * len;
  out[1] = a[1] * len;
  out[2] = a[2] * len;
  return out;
}
/**
 * Calculates the dot product of two vec3's
 *
 * @param {ReadonlyVec3} a the first operand
 * @param {ReadonlyVec3} b the second operand
 * @returns {Number} dot product of a and b
 */

function dot(a, b) {
  return a[0] * b[0] + a[1] * b[1] + a[2] * b[2];
}
/**
 * Computes the cross product of two vec3's
 *
 * @param {vec3} out the receiving vector
 * @param {ReadonlyVec3} a the first operand
 * @param {ReadonlyVec3} b the second operand
 * @returns {vec3} out
 */

function cross(out, a, b) {
  var ax = a[0],
      ay = a[1],
      az = a[2];
  var bx = b[0],
      by = b[1],
      bz = b[2];
  out[0] = ay * bz - az * by;
  out[1] = az * bx - ax * bz;
  out[2] = ax * by - ay * bx;
  return out;
}
/**
 * Alias for {@link vec3.length}
 * @function
 */

var len = length;
/**
 * Perform some operation over an array of vec3s.
 *
 * @param {Array} a the array of vectors to iterate over
 * @param {Number} stride Number of elements between the start of each vec3. If 0 assumes tightly packed
 * @param {Number} offset Number of elements to skip at the beginning of the array
 * @param {Number} count Number of vec3s to iterate over. If 0 iterates over entire array
 * @param {Function} fn Function to call for each vector in the array
 * @param {Object} [arg] additional argument to pass to fn
 * @returns {Array} a
 * @function
 */

var forEach = function () {
  var vec = create$1();
  return function (a, stride, offset, count, fn, arg) {
    var i, l;

    if (!stride) {
      stride = 3;
    }

    if (!offset) {
      offset = 0;
    }

    if (count) {
      l = Math.min(count * stride + offset, a.length);
    } else {
      l = a.length;
    }

    for (i = offset; i < l; i += stride) {
      vec[0] = a[i];
      vec[1] = a[i + 1];
      vec[2] = a[i + 2];
      fn(vec, vec, arg);
      a[i] = vec[0];
      a[i + 1] = vec[1];
      a[i + 2] = vec[2];
    }

    return a;
  };
}();

/**
 * 4 Dimensional Vector
 * @module vec4
 */

/**
 * Creates a new, empty vec4
 *
 * @returns {vec4} a new 4D vector
 */

function create$2() {
  var out = new ARRAY_TYPE(4);

  if (ARRAY_TYPE != Float32Array) {
    out[0] = 0;
    out[1] = 0;
    out[2] = 0;
    out[3] = 0;
  }

  return out;
}
/**
 * Normalize a vec4
 *
 * @param {vec4} out the receiving vector
 * @param {ReadonlyVec4} a vector to normalize
 * @returns {vec4} out
 */

function normalize$1(out, a) {
  var x = a[0];
  var y = a[1];
  var z = a[2];
  var w = a[3];
  var len = x * x + y * y + z * z + w * w;

  if (len > 0) {
    len = 1 / Math.sqrt(len);
  }

  out[0] = x * len;
  out[1] = y * len;
  out[2] = z * len;
  out[3] = w * len;
  return out;
}
/**
 * Perform some operation over an array of vec4s.
 *
 * @param {Array} a the array of vectors to iterate over
 * @param {Number} stride Number of elements between the start of each vec4. If 0 assumes tightly packed
 * @param {Number} offset Number of elements to skip at the beginning of the array
 * @param {Number} count Number of vec4s to iterate over. If 0 iterates over entire array
 * @param {Function} fn Function to call for each vector in the array
 * @param {Object} [arg] additional argument to pass to fn
 * @returns {Array} a
 * @function
 */

var forEach$1 = function () {
  var vec = create$2();
  return function (a, stride, offset, count, fn, arg) {
    var i, l;

    if (!stride) {
      stride = 4;
    }

    if (!offset) {
      offset = 0;
    }

    if (count) {
      l = Math.min(count * stride + offset, a.length);
    } else {
      l = a.length;
    }

    for (i = offset; i < l; i += stride) {
      vec[0] = a[i];
      vec[1] = a[i + 1];
      vec[2] = a[i + 2];
      vec[3] = a[i + 3];
      fn(vec, vec, arg);
      a[i] = vec[0];
      a[i + 1] = vec[1];
      a[i + 2] = vec[2];
      a[i + 3] = vec[3];
    }

    return a;
  };
}();

/**
 * Quaternion
 * @module quat
 */

/**
 * Creates a new identity quat
 *
 * @returns {quat} a new quaternion
 */

function create$3() {
  var out = new ARRAY_TYPE(4);

  if (ARRAY_TYPE != Float32Array) {
    out[0] = 0;
    out[1] = 0;
    out[2] = 0;
  }

  out[3] = 1;
  return out;
}
/**
 * Sets a quat from the given angle and rotation axis,
 * then returns it.
 *
 * @param {quat} out the receiving quaternion
 * @param {ReadonlyVec3} axis the axis around which to rotate
 * @param {Number} rad the angle in radians
 * @returns {quat} out
 **/

function setAxisAngle(out, axis, rad) {
  rad = rad * 0.5;
  var s = Math.sin(rad);
  out[0] = s * axis[0];
  out[1] = s * axis[1];
  out[2] = s * axis[2];
  out[3] = Math.cos(rad);
  return out;
}
/**
 * Performs a spherical linear interpolation between two quat
 *
 * @param {quat} out the receiving quaternion
 * @param {ReadonlyQuat} a the first operand
 * @param {ReadonlyQuat} b the second operand
 * @param {Number} t interpolation amount, in the range [0-1], between the two inputs
 * @returns {quat} out
 */

function slerp(out, a, b, t) {
  // benchmarks:
  //    http://jsperf.com/quaternion-slerp-implementations
  var ax = a[0],
      ay = a[1],
      az = a[2],
      aw = a[3];
  var bx = b[0],
      by = b[1],
      bz = b[2],
      bw = b[3];
  var omega, cosom, sinom, scale0, scale1; // calc cosine

  cosom = ax * bx + ay * by + az * bz + aw * bw; // adjust signs (if necessary)

  if (cosom < 0.0) {
    cosom = -cosom;
    bx = -bx;
    by = -by;
    bz = -bz;
    bw = -bw;
  } // calculate coefficients


  if (1.0 - cosom > EPSILON) {
    // standard case (slerp)
    omega = Math.acos(cosom);
    sinom = Math.sin(omega);
    scale0 = Math.sin((1.0 - t) * omega) / sinom;
    scale1 = Math.sin(t * omega) / sinom;
  } else {
    // "from" and "to" quaternions are very close
    //  ... so we can do a linear interpolation
    scale0 = 1.0 - t;
    scale1 = t;
  } // calculate final values


  out[0] = scale0 * ax + scale1 * bx;
  out[1] = scale0 * ay + scale1 * by;
  out[2] = scale0 * az + scale1 * bz;
  out[3] = scale0 * aw + scale1 * bw;
  return out;
}
/**
 * Creates a quaternion from the given 3x3 rotation matrix.
 *
 * NOTE: The resultant quaternion is not normalized, so you should be sure
 * to renormalize the quaternion yourself where necessary.
 *
 * @param {quat} out the receiving quaternion
 * @param {ReadonlyMat3} m rotation matrix
 * @returns {quat} out
 * @function
 */

function fromMat3(out, m) {
  // Algorithm in Ken Shoemake's article in 1987 SIGGRAPH course notes
  // article "Quaternion Calculus and Fast Animation".
  var fTrace = m[0] + m[4] + m[8];
  var fRoot;

  if (fTrace > 0.0) {
    // |w| > 1/2, may as well choose w > 1/2
    fRoot = Math.sqrt(fTrace + 1.0); // 2w

    out[3] = 0.5 * fRoot;
    fRoot = 0.5 / fRoot; // 1/(4w)

    out[0] = (m[5] - m[7]) * fRoot;
    out[1] = (m[6] - m[2]) * fRoot;
    out[2] = (m[1] - m[3]) * fRoot;
  } else {
    // |w| <= 1/2
    var i = 0;
    if (m[4] > m[0]) i = 1;
    if (m[8] > m[i * 3 + i]) i = 2;
    var j = (i + 1) % 3;
    var k = (i + 2) % 3;
    fRoot = Math.sqrt(m[i * 3 + i] - m[j * 3 + j] - m[k * 3 + k] + 1.0);
    out[i] = 0.5 * fRoot;
    fRoot = 0.5 / fRoot;
    out[3] = (m[j * 3 + k] - m[k * 3 + j]) * fRoot;
    out[j] = (m[j * 3 + i] + m[i * 3 + j]) * fRoot;
    out[k] = (m[k * 3 + i] + m[i * 3 + k]) * fRoot;
  }

  return out;
}
/**
 * Normalize a quat
 *
 * @param {quat} out the receiving quaternion
 * @param {ReadonlyQuat} a quaternion to normalize
 * @returns {quat} out
 * @function
 */

var normalize$2 = normalize$1;
/**
 * Sets a quaternion to represent the shortest rotation from one
 * vector to another.
 *
 * Both vectors are assumed to be unit length.
 *
 * @param {quat} out the receiving quaternion.
 * @param {ReadonlyVec3} a the initial vector
 * @param {ReadonlyVec3} b the destination vector
 * @returns {quat} out
 */

var rotationTo = function () {
  var tmpvec3 = create$1();
  var xUnitVec3 = fromValues(1, 0, 0);
  var yUnitVec3 = fromValues(0, 1, 0);
  return function (out, a, b) {
    var dot$1 = dot(a, b);

    if (dot$1 < -0.999999) {
      cross(tmpvec3, xUnitVec3, a);
      if (len(tmpvec3) < 0.000001) cross(tmpvec3, yUnitVec3, a);
      normalize(tmpvec3, tmpvec3);
      setAxisAngle(out, tmpvec3, Math.PI);
      return out;
    } else if (dot$1 > 0.999999) {
      out[0] = 0;
      out[1] = 0;
      out[2] = 0;
      out[3] = 1;
      return out;
    } else {
      cross(tmpvec3, a, b);
      out[0] = tmpvec3[0];
      out[1] = tmpvec3[1];
      out[2] = tmpvec3[2];
      out[3] = 1 + dot$1;
      return normalize$2(out, out);
    }
  };
}();
/**
 * Performs a spherical linear interpolation with two control points
 *
 * @param {quat} out the receiving quaternion
 * @param {ReadonlyQuat} a the first operand
 * @param {ReadonlyQuat} b the second operand
 * @param {ReadonlyQuat} c the third operand
 * @param {ReadonlyQuat} d the fourth operand
 * @param {Number} t interpolation amount, in the range [0-1], between the two inputs
 * @returns {quat} out
 */

var sqlerp = function () {
  var temp1 = create$3();
  var temp2 = create$3();
  return function (out, a, b, c, d, t) {
    slerp(temp1, a, d, t);
    slerp(temp2, b, c, t);
    slerp(out, temp1, temp2, 2 * t * (1 - t));
    return out;
  };
}();
/**
 * Sets the specified quaternion with values corresponding to the given
 * axes. Each axis is a vec3 and is expected to be unit length and
 * perpendicular to all other specified axes.
 *
 * @param {ReadonlyVec3} view  the vector representing the viewing direction
 * @param {ReadonlyVec3} right the vector representing the local "right" direction
 * @param {ReadonlyVec3} up    the vector representing the local "up" direction
 * @returns {quat} out
 */

var setAxes = function () {
  var matr = create();
  return function (out, view, right, up) {
    matr[0] = right[0];
    matr[3] = right[1];
    matr[6] = right[2];
    matr[1] = up[0];
    matr[4] = up[1];
    matr[7] = up[2];
    matr[2] = -view[0];
    matr[5] = -view[1];
    matr[8] = -view[2];
    return normalize$2(out, fromMat3(out, matr));
  };
}();

function createSVGElement(type) {
  return document.createElementNS('http://www.w3.org/2000/svg', type);
} // /**

var _SHAPE_TO_TAGS, _class, _temp;

var SVGRendererPlugin_1;

var SHAPE_TO_TAGS = (_SHAPE_TO_TAGS = {}, _defineProperty(_SHAPE_TO_TAGS, SHAPE.Rect, 'path'), _defineProperty(_SHAPE_TO_TAGS, SHAPE.Circle, 'circle'), _defineProperty(_SHAPE_TO_TAGS, SHAPE.Ellipse, 'ellipse'), _defineProperty(_SHAPE_TO_TAGS, SHAPE.Image, 'image'), _defineProperty(_SHAPE_TO_TAGS, SHAPE.Group, 'g'), _defineProperty(_SHAPE_TO_TAGS, SHAPE.Line, 'line'), _defineProperty(_SHAPE_TO_TAGS, SHAPE.Polyline, 'polyline'), _defineProperty(_SHAPE_TO_TAGS, SHAPE.Polygon, 'polygon'), _defineProperty(_SHAPE_TO_TAGS, SHAPE.Text, 'text'), _defineProperty(_SHAPE_TO_TAGS, SHAPE.Path, 'path'), _SHAPE_TO_TAGS);
var SVG_ATTR_MAP = {
  opacity: 'opacity',
  fillStyle: 'fill',
  fill: 'fill',
  fillOpacity: 'fill-opacity',
  strokeStyle: 'stroke',
  strokeOpacity: 'stroke-opacity',
  stroke: 'stroke',
  r: 'r',
  rx: 'rx',
  ry: 'ry',
  width: 'width',
  height: 'height',
  // x1: 'x1',
  // x2: 'x2',
  // y1: 'y1',
  // y2: 'y2',
  lineCap: 'stroke-linecap',
  lineJoin: 'stroke-linejoin',
  lineWidth: 'stroke-width',
  lineDash: 'stroke-dasharray',
  lineDashOffset: 'stroke-dashoffset',
  miterLimit: 'stroke-miterlimit',
  font: 'font',
  fontSize: 'font-size',
  fontStyle: 'font-style',
  fontVariant: 'font-variant',
  fontWeight: 'font-weight',
  fontFamily: 'font-family',
  letterSpacing: 'letter-spacing',
  startArrow: 'marker-start',
  endArrow: 'marker-end',
  // path: 'd',
  class: 'class',
  id: 'id',
  style: 'style',
  preserveAspectRatio: 'preserveAspectRatio',
  visibility: 'visibility'
};
var SVGRendererPlugin = SVGRendererPlugin_1 = (_temp = _class = /*#__PURE__*/function () {
  function SVGRendererPlugin() {
    _classCallCheck(this, SVGRendererPlugin);

    this.camera = void 0;
    this.contextService = void 0;
    this.sceneGraphService = void 0;
    this.elementRendererFactory = void 0;
  }

  _createClass(SVGRendererPlugin, [{
    key: "apply",
    value: function apply(renderingService) {
      var _this = this;

      renderingService.hooks.mounted.tap(SVGRendererPlugin_1.tag, function (object) {
        var entity = object.getEntity(); // create svg element

        var svgElement = entity.addComponent(ElementSVG);
        var type = SHAPE_TO_TAGS[object.nodeType];

        if (type) {
          var $groupEl;
          var $el = createSVGElement(type);
          $el.id = entity.getName();

          if (type !== 'g') {
            $groupEl = createSVGElement('g');
            $groupEl.appendChild($el);
          } else {
            $groupEl = $el;
          }

          svgElement.$el = $el;
          svgElement.$groupEl = $groupEl;

          var $parentGroupEl = object.parentNode && object.parentNode.getEntity().getComponent(ElementSVG).$groupEl || _this.contextService.getContext();

          if ($parentGroupEl) {
            $parentGroupEl.appendChild($groupEl);
          }
        }
      });
      renderingService.hooks.unmounted.tap(SVGRendererPlugin_1.tag, function (object) {
        object.getEntity().removeComponent(ElementSVG, true);
      });
      renderingService.hooks.render.tap(SVGRendererPlugin_1.tag, function (object) {
        var _entity$getComponent, _entity$getComponent2;

        var $namespace = _this.contextService.getDomElement();

        if ($namespace) {
          // @ts-ignore
          _this.applyTransform($namespace, _this.camera.getOrthoMatrix());
        }

        var entity = object.getEntity();
        var $el = (_entity$getComponent = entity.getComponent(ElementSVG)) === null || _entity$getComponent === void 0 ? void 0 : _entity$getComponent.$el;
        var $groupEl = (_entity$getComponent2 = entity.getComponent(ElementSVG)) === null || _entity$getComponent2 === void 0 ? void 0 : _entity$getComponent2.$groupEl;

        if ($el && $groupEl) {
          var nodeType = object.nodeType,
              attributes = object.attributes; // apply local RTS transformation to <group> wrapper

          _this.applyTransform($groupEl, object.getLocalTransform());

          $el.setAttribute('fill', 'none');

          if (nodeType === SHAPE.Image) {
            $el.setAttribute('preserveAspectRatio', 'none');
          } // apply attributes


          for (var name in attributes) {
            _this.updateAttribute(entity, name, "".concat(attributes[name]));
          } // generate path


          var elementRenderer = _this.elementRendererFactory(nodeType);

          if (elementRenderer) {
            elementRenderer.apply($el, attributes);
          }

          _this.reorderChildren($groupEl, object.children || []); // finish rendering, clear dirty flag


          var renderable = entity.getComponent(Renderable);
          renderable.dirty = false;
        }
      });
      renderingService.hooks.attributeChanged.tap(SVGRendererPlugin_1.tag, function (object, name, value) {
        var entity = object.getEntity();

        if (name === 'zIndex') {
          var _object$parentNode, _parentEntity$getComp;

          var parent = object.parentNode;
          var parentEntity = (_object$parentNode = object.parentNode) === null || _object$parentNode === void 0 ? void 0 : _object$parentNode.getEntity();
          var $groupEl = parentEntity === null || parentEntity === void 0 ? void 0 : (_parentEntity$getComp = parentEntity.getComponent(ElementSVG)) === null || _parentEntity$getComp === void 0 ? void 0 : _parentEntity$getComp.$groupEl;

          var children = _toConsumableArray((parent === null || parent === void 0 ? void 0 : parent.children) || []);

          if ($groupEl) {
            _this.reorderChildren($groupEl, children);
          }
        }

        _this.updateAttribute(entity, name, value);
      });
    }
  }, {
    key: "reorderChildren",
    value: function reorderChildren($groupEl, children) {
      // need to reorder parent's children
      children.sort(this.sceneGraphService.sort); // create empty fragment

      var fragment = document.createDocumentFragment();
      children.forEach(function (child) {
        var $el = child.getEntity().getComponent(ElementSVG).$groupEl;

        if ($el) {
          fragment.appendChild($el);
        }
      });
      $groupEl.appendChild(fragment);
    }
  }, {
    key: "applyTransform",
    value: function applyTransform($el, transform) {
      var translation = getTranslation(create$1(), transform);
      var scaling = getScaling(create$1(), transform);
      var rotation = getRotation(create$3(), transform);

      var _translation = _slicedToArray(translation, 2),
          x = _translation[0],
          y = _translation[1];

      var _scaling = _slicedToArray(scaling, 2),
          scaleX = _scaling[0],
          scaleY = _scaling[1];

      var _getEuler = getEuler(create$1(), rotation),
          _getEuler2 = _slicedToArray(_getEuler, 3),
          ex = _getEuler2[0],
          ey = _getEuler2[1],
          ez = _getEuler2[2]; // gimbal lock at 90 degrees


      var rts = fromRotationTranslationScale(ex || ez, x, y, scaleX, scaleY); // TODO: use proper precision avoiding too long string in `transform`
      // @see https://developer.mozilla.org/en-US/docs/Web/API/Canvas_API/Tutorial/Transformations

      $el.setAttribute('transform', "matrix(".concat(rts[0], ",").concat(rts[1], ",").concat(rts[3], ",").concat(rts[4], ",").concat(rts[6], ",").concat(rts[7], ")"));
    }
  }, {
    key: "updateAttribute",
    value: function updateAttribute(entity, name, value) {
      if (SVG_ATTR_MAP[name]) {
        // update `visibility` on <group>
        if (name === 'visibility') {
          var _entity$getComponent3, _entity$getComponent4;

          (_entity$getComponent3 = entity.getComponent(ElementSVG)) === null || _entity$getComponent3 === void 0 ? void 0 : (_entity$getComponent4 = _entity$getComponent3.$groupEl) === null || _entity$getComponent4 === void 0 ? void 0 : _entity$getComponent4.setAttribute(SVG_ATTR_MAP[name], "".concat(value));
        } else {
          var _entity$getComponent5, _entity$getComponent6;

          (_entity$getComponent5 = entity.getComponent(ElementSVG)) === null || _entity$getComponent5 === void 0 ? void 0 : (_entity$getComponent6 = _entity$getComponent5.$el) === null || _entity$getComponent6 === void 0 ? void 0 : _entity$getComponent6.setAttribute(SVG_ATTR_MAP[name], "".concat(value));
        }
      }
    }
  }]);

  return SVGRendererPlugin;
}(), _class.tag = 'SVGRendererPlugin', _temp);

__decorate([inject(Camera), __metadata("design:type", Camera)], SVGRendererPlugin.prototype, "camera", void 0);

__decorate([inject(ContextService), __metadata("design:type", Object)], SVGRendererPlugin.prototype, "contextService", void 0);

__decorate([inject(SceneGraphService), __metadata("design:type", SceneGraphService)], SVGRendererPlugin.prototype, "sceneGraphService", void 0);

__decorate([inject(ElementRendererFactory), __metadata("design:type", Function)], SVGRendererPlugin.prototype, "elementRendererFactory", void 0);

SVGRendererPlugin = SVGRendererPlugin_1 = __decorate([injectable()], SVGRendererPlugin);

world.registerComponent(ElementSVG);
var containerModule = new ContainerModule(function (bind, unbind, isBound, rebind) {
  /**
   * register shape renderers
   */
  bind(RectRenderer).toSelf().inSingletonScope();
  bind(ImageRenderer).toSelf().inSingletonScope();
  bind(LineRenderer).toSelf().inSingletonScope();
  bind(PolylineRenderer).toSelf().inSingletonScope();
  bind(TextRenderer).toSelf().inSingletonScope();
  bind(PathRenderer).toSelf().inSingletonScope();
  bind(ElementRendererFactory).toFactory(function (ctx) {
    return function (tagName) {
      if (tagName === SHAPE.Rect) {
        return ctx.container.get(RectRenderer);
      } else if (tagName === SHAPE.Image) {
        return ctx.container.get(ImageRenderer);
      } else if (tagName === SHAPE.Line) {
        return ctx.container.get(LineRenderer);
      } else if (tagName === SHAPE.Polyline || tagName === SHAPE.Polygon) {
        return ctx.container.get(PolylineRenderer);
      } else if (tagName === SHAPE.Text) {
        return ctx.container.get(TextRenderer);
      } else if (tagName === SHAPE.Path) {
        return ctx.container.get(PathRenderer);
      }

      return null;
    };
  });
  bind(SVGRendererPlugin).toSelf().inSingletonScope();
  bind(RenderingPluginContribution).toService(SVGRendererPlugin);
});

export { ElementSVG, containerModule };
